//
// Created by tunahan on 10.02.20.
//

#include <catch2/catch.hpp>
#include <daemon/backend/metadata/db.hpp>
#include <daemon/classes/fs_data.hpp>
#include <global/path_util.hpp>
#include <random>
#include <pthread.h>
#include <thread>
#include <fmt/format.h>
#include <set>
#include <filesystem>
#include <string>
#include <boost/filesystem.hpp>

// nach auflistugn/beschreibungsmöglichkeiten schauen
// -l -t für auflistung
// --list-test-names-only

std::string gen_random(const int len, bool key) {
    std::string s;
    s.resize(len);
    static const char alphanum[] =
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "abcdefghijklmnopqrstuvwxyz";

    static const char numbers[] =  "123456789";
    if (key) {
        s[0] = '/';
        for (int i = 1; i < len; ++i) {
            s[i] = alphanum[std::rand() % (sizeof(alphanum) - 1)];
        }
        s[len] = '\0';
    }else{
        if(len < 6) return s;
        for (int i = 0; i < len-5; ++i) {
            s[i] = numbers[std::rand() % (sizeof(numbers) - 1)];
        }
        s[len-5] = '|';
        for (int i = len-4; i < len; ++i) {
            s[i] = numbers[std::rand() % (sizeof(numbers) - 1)];
        }

        s[len-1] = '|';
    }
    return s;
}
// BDD STYLE

int i = 0;
SCENARIO( "RDB - Input single Key", "[RocksDB]") {
    GIVEN("Open a db") {

        // when open db
        std::string metadata_path = "/tmp/db_test" + std::to_string(i);
        gkfs::metadata::MetadataDB db = gkfs::metadata::MetadataDB(metadata_path);
        i++;
        WHEN("Try to input key") {
            std::string key = "/a";
            std::string val = "453|123|";

            db.put(key, val);

            THEN("Key and value in DB") {
                auto res = db.get(key);
                REQUIRE(res == val);
            }
        }

        boost::filesystem::remove_all(metadata_path);

    }
}

SCENARIO( "RDB - Insert 1 key with two values", "[RocksDB]") {
    GIVEN("Open a db") {

        // when open db
        std::string metadata_path = "/tmp/db_test" + std::to_string(i);
        gkfs::metadata::MetadataDB db = gkfs::metadata::MetadataDB(metadata_path);
        i++;

        WHEN("RocksDB 1 key two values - returns first value"){
            std::string key = "/a";
            std::string val = "453|123|";
            std::string val2 = "453|124|";

            THEN("First value returns for key") {
                db.put(key, val);
                db.put(key, val2);

                auto res = db.get(key);
                REQUIRE(res == val);
            }
        }

        boost::filesystem::remove_all(metadata_path);

    }
}


SCENARIO( "RDB - delete existing Key", "[RocksDB]") {
    GIVEN("Open a db") {

        // when open db
        std::string metadata_path = "/tmp/db_test" + std::to_string(i);
        gkfs::metadata::MetadataDB db = gkfs::metadata::MetadataDB(metadata_path);
        i++;

        WHEN("RocksDB delete inserted key") {
            std::string metadata_path = "/home/tunahan/rocksdb";

            std::string key = "/";
            std::string val = "453|123|";
            THEN("Key not in DB anymore") {
                db.put(key, val);
                db.remove(key);
            }
        }


        boost::filesystem::remove_all(metadata_path);

    }
}


SCENARIO( "RDB - Get deleted Key", "[RocksDB]") {
    GIVEN("Open a db") {

        // when open db
        std::string metadata_path = "/tmp/db_test" + std::to_string(i);
        gkfs::metadata::MetadataDB db = gkfs::metadata::MetadataDB(metadata_path);
        i++;

        WHEN("RocksDB get deleted key") {
            std::string metadata_path = "/home/tunahan/rocksdb";

            std::string key = "/";
            std::string val = "453|123|";
            THEN(" requie return = Nor Found: ") {
                db.put(key, val);
                db.remove(key);
                REQUIRE_THROWS(db.get(key), "Not Found: ");
            }
        }

        boost::filesystem::remove_all(metadata_path);

    }

}

SCENARIO( "RDB - update existing Key", "[RocksDB]") {
    GIVEN("Open a db") {

        // when open db
        std::string metadata_path = "/tmp/db_test" + std::to_string(i);
        gkfs::metadata::MetadataDB db = gkfs::metadata::MetadataDB(metadata_path);
        i++;



        WHEN("RocksDB update key") {
            std::string metadata_path = "/home/tunahan/rocksdb";

            std::string key = "/";
            std::string key2 = "/a";
            std::string val = "453|123|";
            std::string val2 = "453|151|";
            THEN("require old key deleted and new key inserted") {
                db.put(key, val);
                db.update(key, key2, val2);
                auto res2 = db.get(key2);
                REQUIRE_THROWS(db.get(key), "Not Found: ");
                REQUIRE(res2 == val2);
            }
        }

        boost::filesystem::remove_all(metadata_path);


    }
}


SCENARIO( "RDB - Bulk Put/Get - single core", "[RocksDB]") {
    GIVEN("Open a db") {

        // when open db
        std::string metadata_path = "/tmp/db_test" + std::to_string(i);
        gkfs::metadata::MetadataDB db = gkfs::metadata::MetadataDB(metadata_path);
        i++;

        WHEN("Bulk Insert/Get - single core") {


            const int number_Entries = 100;
            const int len_strings = 10;
            std::set<std::string> keyList;
            std::vector<std::string> valList;
            for (int j = 0; j < number_Entries; j++) {
                std::string temp;
                temp = gen_random(len_strings, false);
                valList.push_back(temp);
                //std::cout << temp << "\n";
                temp = gen_random(len_strings, true);
                keyList.insert(temp);
                //std::cout << temp << "\n";
            }

            THEN("require all successfull") {

                auto keys = keyList.begin();
                for (int j = 0; j < number_Entries; j++) {
                    auto key = *(keys++);
                    db.put(key, valList[j]);
                    auto res = db.get(key);
                    REQUIRE(res == valList[j]);
                }
            }
        }

        boost::filesystem::remove_all(metadata_path);


    }
}



SCENARIO( "RDB - Multithreaded Bulk Insert and Get", "[RocksDB]") {
    GIVEN("Open a db") {

        // when open db
        std::string metadata_path = "/tmp/db_test" + std::to_string(i);
        gkfs::metadata::MetadataDB db = gkfs::metadata::MetadataDB(metadata_path);
        i++;

        WHEN("Multithreaded Bulk Insert and Get") {

            const int number_Entries = 1000;


            std::vector<std::thread> threads;

            for (int i = 0; i < 8; ++i) {
                threads.emplace_back([&]() {
                    const int len_strings = 10;
                    std::set<std::string> keyList;
                    std::vector<std::string> valList;
                    for (int i = 0; i < number_Entries; i++) {
                        std::string temp;
                        temp = gen_random(len_strings, false);
                        valList.push_back(temp);
                        temp = gen_random(len_strings, true);
                        keyList.insert(temp);
                    }
                    auto keys = keyList.begin();
                    for (int i = 0; i < number_Entries; i++) {
                        auto key = *(keys++);
                        db.put(key, valList[i]);
                        auto res = db.get(key);
                        REQUIRE(res == valList[i]);
                    }
                });
            }

            for (auto &t : threads) { t.join(); }

        }
        boost::filesystem::remove_all(metadata_path);


    }
}
SCENARIO( "RDB - Merge Operator", "[RocksDB]") {
    GIVEN("Open a db") {

        // when open db
        std::string metadata_path = "/tmp/db_test" + std::to_string(i);
        gkfs::metadata::MetadataDB db = gkfs::metadata::MetadataDB(metadata_path);
        i++;
        WHEN("Merge Operation") {

            std::string key = "/a";
            std::string val = "4523|123|";

            THEN("increase size of key") {
                db.put(key, val);
                db.increase_size(key, 10, true);
            }
        }
        boost::filesystem::remove_all(metadata_path);

    }

}

